//Copyright © mSIGNIA, Incorporated, 2007.  All rights reserved.
//
//Software is protected by one or more of U.S. Patent No. 9,559,852, 9294448, 8,817,984,
//international patents and others pending. For more information see www.mSIGNIA.com.  User agrees
//that they will not them self, or through any affiliate, agent or other third-party, entity or
//other business structure remove, alter, cover or obfuscate any copyright notices or other
//proprietary rights notices of mSIGNIA or its licensors.  User agrees that they will not them
//self, or through any affiliate, agent or other third party, entity or other business structure
//(a) reproduce, sell, lease, license or sublicense this software or any part thereof, (b)
//decompile, disassemble, re-program, reverse engineer or otherwise attempt to derive or modify
//this software in whole or in part, (c) write or develop any derivative software or any other
//software program based upon this software, (d) provide, copy, transmit, disclose, divulge, or
//make available to, or permit use of this software by any third party or entity or machine without
//software owner's prior written consent, (e) circumvent or disable any security or other
//technological features or measures used by this software.
//
//  WebService.swift
//  ThreeDSStoreDemo
//
//  Created by Drew Pitchford on 12/7/18.
//  Copyright © 2018 mSignia. All rights reserved.
//

import Foundation

/*
 @param data - The json data returned by the server
 @param urlResponse - The response information
 @param error - The error returned by the server
*/

class WebService {
    
    // MARK: - Singleton
    static var shared: WebService {
        
        guard let baseURL = URL(string: BuildEnvironment.baseURL) else {
            fatalError("Pass a base URL to Web Service in order to use it")
        }
        
        let webService = WebService(baseURL: baseURL)
        return webService
    }
    
    // MARK: - Typealises
    typealias WebCompletionHandler = (_: Data?, _: URLResponse?, _: Error?) -> Void
    
    // MARK: - HTTP Methods
    private enum HTTPMethod: String {
        
        case post = "POST"
        case get = "GET"
        case put = "PUT"
        case delete = "DELETE"
    }
    
    // MARK: - Properties
    var baseURL: URL
    private var session: URLSession {
        let config = URLSessionConfiguration.default
        return URLSession(configuration: config)
    }
    var headers: [String: String] {
        let headers = ["Content-Type": "application/json", "Accept": "application/json", "Accept-Encoding": "gzip"]
        return headers
    }
    private var requestCreationError = NSError(domain: "Web Service Request Error", code: -1, userInfo: [NSLocalizedDescriptionKey: "Request could not be formed"])
    
    // MARK: -Inits
    init(baseURL: URL) {
        self.baseURL = baseURL
    }
    
    // MARK: - Requests
    private func createRequest(with method: HTTPMethod, forEndpoint endpoint: String, and queryItems: [URLQueryItem]? = nil) -> URLRequest? {
        var urlComponents = URLComponents(string: baseURL.absoluteString + endpoint)
        urlComponents?.queryItems = queryItems
        guard let fullURL = urlComponents?.url else {
            return nil
        }
        
        var request = URLRequest(url: fullURL)
        request.httpMethod = method.rawValue
        request.timeoutInterval = 60
        request.allHTTPHeaderFields = headers
        return request
    }
}

// MARK: - Custom HTTP calls
extension WebService {
    func msPost(_ endpoint: String, params: Data?, completion: @escaping WebCompletionHandler) {
        httpRequest(forEndpoint: endpoint, with: .post, params: params, completion: completion)
    }
    
    func msPut(_ endpoint: String, params: Data?, completion: @escaping WebCompletionHandler) {
        httpRequest(forEndpoint: endpoint, with: .put, params: params, completion: completion)
    }
    
    func msGet(_ endpoint: String, queryItems: [URLQueryItem]? = nil, completion: @escaping WebCompletionHandler) {
        guard let request = createRequest(with: .get, forEndpoint: endpoint, and: queryItems) else {
            completion(nil, nil, requestCreationError)
            return
        }
            
        session.dataTask(with: request) { (data, urlResponse, error) in
            DispatchQueue.main.async {
                self.handleResponse(in: data, with: urlResponse, and: error, completion: completion)
            }
        }.resume()
    }
    
    
    func msDelete(_ endpoint: String, params: Data? = nil, completion: @escaping WebCompletionHandler) {
        httpRequest(forEndpoint: endpoint, with: .delete, params: params, completion: completion)
    }
    
    // MARK: - HTTP Request Methods
    private func httpRequest(forEndpoint endpoint: String, with method: HTTPMethod, params: Data?, completion: @escaping WebCompletionHandler) {
        guard var request = createRequest(with: method, forEndpoint: endpoint) else {
            // do something here like call the completion with an error
            completion(nil, nil, requestCreationError)
            return
        }
        
        request.httpBody = params
        session.dataTask(with: request) { (data, urlResponse, error) in
            DispatchQueue.main.async {
                self.handleResponse(in: data, with: urlResponse, and: error, completion: completion)
            }
        }.resume()
    }
    
    // MARK: - 200 Response Error Helpers
    func handleResponse(in data: Data?, with urlResponse: URLResponse?, and error: Error?, completion: @escaping WebCompletionHandler) {
        completion(data, urlResponse, error)
    }
}
