//
//  HomeViewController.swift
//  ThreeDS-Sample
//
//  Created by Drew Pitchford on 5/14/20.
//  Copyright © 2020 mSignia. All rights reserved.
//

import UIKit
import CoreLocation
import uSDK

class HomeTableViewController: UITableViewController, AppInfoDetailable {
    
    // MARK: - IBOutlets
    @IBOutlet weak var createTransactionButton: UIButton!
    @IBOutlet weak var acctNumberTextField: UITextField!
    @IBOutlet weak var emailTextField: UITextField!
    @IBOutlet weak var pasteboardButton: UIButton!
    @IBOutlet weak var backendURLTextField: UITextField!
    @IBOutlet weak var versionLabel: UILabel!
    
    
    // MARK: - Local Constants
    private let defaultAcctNumber = "4254260145873316"
    private let defaultEmail = "qa@msignia.com"
    private let defaultUIDelay = 0.1
    
    // MARK: - Properties
    private var configParams: UConfigParameters {
        let params = UConfigParameters()
        let ec = UDirectoryServer(dsid: Constants.DS.ecName, publicKey: Constants.DS.ecKey, keyID: Constants.DS.ecKeyID, dsCACertificate: Constants.DS.cert, providerName: "mSIGINIA EC", dsLogo: UIImage(named: "cb_card"))
        let rsa = UDirectoryServer(dsid: Constants.DS.rsaName, publicKey: Constants.DS.rsaKey, keyID: Constants.DS.rsaKeyID, dsCACertificate: Constants.DS.cert, providerName: "mSIGNIA RSA", dsLogo: UIImage(named: "mastercard"))
        
        params.add(ec)
        params.add(rsa)
        return params
    }
    private var purchaseInfo: PurchaseInformation {
        return PurchaseInformation(email: emailTextField.text ?? defaultEmail, cardID: acctNumberTextField.text ?? defaultAcctNumber, purchaseAmount: "100", purchaseCurrency: "840", purchaseExponent: "2", purchaseDate: "20200608094835")
    }
    private var licenseKey: String {
        return BuildEnvironment.licenseKey
    }
    private var lastEnteredEmailAddress: String {
        set {
            UserDefaults.standard.set(!newValue.isEmpty ? newValue : defaultEmail, forKey: Constants.UserDefaults.Keys.email)
        }
        
        get {
            UserDefaults.standard.string(forKey: Constants.UserDefaults.Keys.email) ?? defaultEmail
        }
    }
    private var spinner: UIActivityIndicatorView?
    private var trans: UTransaction! // Never do implicity unwrapped optionals - this is just the fastest way to demo
    private var ares: AuthenticationResponse! // Same as above
    var lastThreeDSServerTransID: String?
    
    // MARK: - Lifecycle
    override func viewDidLoad() {
        super.viewDidLoad()
        
        setUpUI()
    }
    
    override func viewDidAppear(_ animated: Bool) {
        super.viewDidAppear(animated)
    
        LocationService.shared.delegate = self
        LocationService.shared.requestPermission()
    }
    
    // MARK: - Setup
    private func setUpUI() {
        // Buttons
        createTransactionButton.layer.cornerRadius = Constants.UI.Radii.default
        pasteboardButton.layer.cornerRadius = Constants.UI.Radii.default
        
        // Text Fields
        acctNumberTextField.text = defaultAcctNumber
        acctNumberTextField.layer.cornerRadius = Constants.UI.Radii.default
        
        emailTextField.layer.cornerRadius = Constants.UI.Radii.default
        emailTextField.text = lastEnteredEmailAddress
        
        backendURLTextField.layer.cornerRadius = Constants.UI.Radii.default
        backendURLTextField.text = WebService.shared.baseURL.absoluteString
        
        // Activity Indicator
        if #available(iOS 13.0, *) {
            spinner = UIActivityIndicatorView(style: .large)
        }
        else {
            spinner = UIActivityIndicatorView(style: .whiteLarge)
        }
        
        spinner?.center = CGPoint(x: UIScreen.main.bounds.size.width / 2, y: UIScreen.main.bounds.size.height / 2)
        if let spinner = spinner {
            view.addSubview(spinner)
        }
        
        // Version Label
        versionLabel.text = "app: \(fetchAppVersion() ?? "N/A"), usdk: \(fetchUSDKVersion() ?? "N/A")"
    }
    
    // MARK: - Helpers
    private func copyStringToPasteboard(string: String?) {
        if let str = string {
            UIPasteboard.general.string = str
        }
    }
    
    private func areTextFieldsValid() -> (emailIsValid: Bool, acctNumberIsValid: Bool) {
        var emailIsValid = false
        if let email = emailTextField.text,
            !email.isEmpty {
            emailIsValid = true
        }
        
        var acctNumberIsValid = false
        if let acctNumber = acctNumberTextField.text,
            !acctNumber.isEmpty {
            acctNumberIsValid = true
        }
        
        return (emailIsValid, acctNumberIsValid)
    }
    
    // MARK: - IBActions
    @IBAction func createTransactionWasPressed(_ sender: UIButton) {
        let textFieldValidity = areTextFieldsValid()
        if textFieldValidity.emailIsValid &&
            textFieldValidity.acctNumberIsValid {
            spinner?.startAnimating()
            view.endEditing(true)
            authenticateUser()
        }
        else if !textFieldValidity.emailIsValid {
            showAlert(withTitle: "Missing Value", message: "Please enter an email address")
        }
        else if !textFieldValidity.acctNumberIsValid {
            showAlert(withTitle: "Missing Value", message: "Please enter an account number")
        }
    }
    
    @IBAction func copyLastTransIDButtonPressed(_ sender: UIButton) {
        copyStringToPasteboard(string: lastThreeDSServerTransID)
    }
    
    // MARK: - Actions
    private func authenticateUser() {
        if let acctNumber = acctNumberTextField.text {
            var headers = WebService.shared.headers
            headers["email"] = emailTextField.text
            let authSpec = UAuthenticationSpec(cardId: acctNumber, purchaseInfo: purchaseInfo, isPaymentAuthentication: true, supportedVersionsURL: "\(WebService.shared.baseURL)/api/v3/supported-versions", authURL: "\(WebService.shared.baseURL)/api/v2/auth", authorizationHeaders: headers)
            UThreeDS2ServiceImpl.shared().authenticate(spec: authSpec, delegate: self, currentViewController: self)
        }
        else {
            showAlert(message: "Account Number was not found. Did you type something into the text field above?")
        }
    }
    
    private func initSDK() {
        SDKThemeManager.shared.setDefaultTheme()
        let customization = SDKThemeManager.shared.customization
        let initSpec = UInitSpec(licenseKey: licenseKey, locale: "us", uiCustomization: customization, configParameters: configParams)
        UThreeDS2ServiceImpl.shared().initialize(spec: initSpec) { (error) in
            DispatchQueue.main.async {
                if let err = error {
                    print("Error initializing")
                    self.showAlert(for: err)
                }
                else {
                    self.showAlert(withTitle: "SDK Initialization", message: "Success!")
                }
            }
        }
    }
}

// MARK: - UITextFieldDelegate methods
extension HomeTableViewController: UITextFieldDelegate {
    func textFieldDidEndEditing(_ textField: UITextField) {
        if let text = textField.text {
            switch textField {
            case emailTextField:
                lastEnteredEmailAddress = text
                
            case backendURLTextField:
                BuildEnvironment.baseURL = text
                
            default:
                return
            }
        }
    }
    
    func textFieldShouldReturn(_ textField: UITextField) -> Bool {
        switch textField {
        case backendURLTextField:
            textField.resignFirstResponder()
        default:
            textField.nextField?.becomeFirstResponder()
        }
        
        return true
    }
}

// MARK: - LocationServiceDelegate methods
extension HomeTableViewController: LocationServiceDelegate {
    func authorizationStatusDidChange(to status: CLAuthorizationStatus) {
        self.initSDK()
    }
}

// MARK: - UAuthenticationDelegate methods
extension HomeTableViewController: UAuthenticationDelegate {
    func authenticated(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + defaultUIDelay) {
            self.spinner?.stopAnimating()
            self.lastThreeDSServerTransID = threeDSServerTransID
            self.showAlert(withTitle: "Transaction Authenticated", message: "3DS Server TransID \(threeDSServerTransID)")
            print("User authenticated! \(threeDSServerTransID) \(status)")
        }
    }
    
    func notAuthenticated(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + defaultUIDelay) {
            self.spinner?.stopAnimating()
            self.lastThreeDSServerTransID = threeDSServerTransID
            self.showAlert(withTitle: "Transaction Not Authenticated", message: "3DS Server TransID \(threeDSServerTransID)")
            print("User not authenticated! \(threeDSServerTransID) \(status)")
        }
    }
    
    func decoupledAuthBeingPerformed(_ threeDSServerTransID: String, status: UTransStatus) {
        DispatchQueue.main.asyncAfter(deadline: .now() + defaultUIDelay) {
            self.spinner?.stopAnimating()
            self.lastThreeDSServerTransID = threeDSServerTransID
            self.showAlert(withTitle: "Decoupled Auth Being Performed", message: "3DS Server TransID \(threeDSServerTransID)")
            print("User decoupled-authed! \(threeDSServerTransID) \(status)")
        }
    }
    
    func cancelled(_ threeDSServerTransID: String) {
        DispatchQueue.main.asyncAfter(deadline: .now() + defaultUIDelay) {
            self.spinner?.stopAnimating()
            self.lastThreeDSServerTransID = threeDSServerTransID
            self.showAlert(withTitle: "Transaction Cancelled", message: "3DS Server TransID \(threeDSServerTransID)")
            print("Authentication cancelled! \(threeDSServerTransID)")
        }
    }
    
    func error(_ threeDSServerTransID: String, error: Error) {
        DispatchQueue.main.asyncAfter(deadline: .now() + defaultUIDelay) {
            self.spinner?.stopAnimating()
            self.lastThreeDSServerTransID = threeDSServerTransID
            self.showAlert(for: error)
            print("Error occurred: \(error)")
        }
    }
}

// MARK: - Constants
private struct Constants {
    struct UI {
        struct Radii {
            static let `default`: CGFloat = 10
        }
    }
    
    struct UserDefaults {
        struct Keys {
            static let email = "ud_email_key"
        }
    }
    
    struct DS {
        static let ecName = "MSIGNIA_MOCK_DS_EC"
        static let ecKeyID = "de7655be-3d6f-4802-9d68-c36a9d23e71c"
        static let ecKey = "{\"kty\":\"EC\",\"crv\":\"P-256\",\"x\":\"7YTOG_nT8NtzC7sgQDEEsbkFX6VEXdzXgpoU0FD7gRY\",\"y\":\"Th1ej6HF0K9yVg3DM0K6zQrxERBSoX597LqadlIq77o\"}"
        static let rsaName = "MSIGNIA_MOCK_DS_RSA"
        static let rsaKeyID = "de7655be-3d6f-4802-9d68-c36a9d23e71c"
        static let rsaKey = "{\"kty\":\"RSA\",\"e\":\"AQAB\",\"kid\":\"de7655be-3d6f-4802-9d68-c36a9d23e71c\",\"n\":\"4Uko-Z2ESFV5IUXunK08ouSziFz_OfuRR9H4woBdMbAeyAq0-V5o36tCtiNVYgF2Oi1jBgmEIxIMH9y-W_Mn0OoiwzV7eQj4tc0q5RM9RzYEMoRCK-0YU4nFzRlDSFQOIPEkEWihPHujyY32qBDyMsl0ctlq5EN-F77u9wCvoxU9OLM3dlAE42dNrbrHWOnpeWg41kk63AUowZlA7QE52Tgd_hyaHAy4oQtgp1_-l-SPIAKIoAF-5Ibiss4Y1Wah1RSvDz69UMgHoqc_cvWG9kTPMtkFoIwW8ta2jhGzvhoNf2X1o2excBLKlWT3hIwGfnw8oYRDlINTZIuTGLc0cw\"}"
        
        static let cert = """
        -----BEGIN CERTIFICATE-----
        MIID4zCCAsugAwIBAgIBATANBgkqhkiG9w0BAQsFADB7MQswCQYDVQQGEwJVUzEL
        MAkGA1UECBMCVE4xEDAOBgNVBAoTB01TSUdOSUExDDAKBgNVBAsTA1JORDEXMBUG
        A1UEAxMOTVNJR05JQSBSTkQgQ0ExJjAkBgkqhkiG9w0BCQEWF3BhdmxvLmx5c292
        QG1zaWduaWEuY29tMB4XDTE5MTExMzA3MzYwMFoXDTI5MTExMzA3MzYwMFowezEL
        MAkGA1UEBhMCVVMxCzAJBgNVBAgTAlROMRAwDgYDVQQKEwdNU0lHTklBMQwwCgYD
        VQQLEwNSTkQxFzAVBgNVBAMTDk1TSUdOSUEgUk5EIENBMSYwJAYJKoZIhvcNAQkB
        FhdwYXZsby5seXNvdkBtc2lnbmlhLmNvbTCCASIwDQYJKoZIhvcNAQEBBQADggEP
        ADCCAQoCggEBAPSn+PdwUVXYq5IQN+vrcxM3WfR5Xpgb3OpIRkka863duH0j3Uh/
        poaFxnp/j0dGbefYaKaXGZSFX/E3FWaAFGKHV2Lbt4uxYkJx4Fa2gXsBxqdOd4Li
        aqSRSZXPtYMqoMbCOQj4lqtthBNe5BDX6/1HDJELLo8sxnIyfwxgbome3s9e6Rpz
        p+M9XgOtfyuSzPhzY0bWeBrHDzPcKkxQgKZeywlgApCPPfoYEvBDfzGbNns4yA+v
        sqYGDcctplQceUG6EFN68qbmy3CzV8zQMHFkHqIJ/l1V5RlGT9b5QhNVhu4MJo//
        W8LaA6vdyyqPnS7Hbvp6JklkMVxJJhWl0t8CAwEAAaNyMHAwDwYDVR0TAQH/BAUw
        AwEB/zAdBgNVHQ4EFgQU443mM7cesLlNiq01jfVfyJ7Rh1gwCwYDVR0PBAQDAgEG
        MBEGCWCGSAGG+EIBAQQEAwIABzAeBglghkgBhvhCAQ0EERYPeGNhIGNlcnRpZmlj
        YXRlMA0GCSqGSIb3DQEBCwUAA4IBAQBOY7ETJ0MyrQ44MYNbkGI7fIij5oxM0In0
        Z4ZT5NCA1p3fJP/PV8oxONE+7ix1GlCw8RR59L2VQGkg4I2vSj0DHqpuPf3u86Fm
        gZtkhcF41dlWSic1B5oaA610iglITgu/gyZrC+r6RHGIKfyLX+u8YGc6WfI8g/ED
        aVQcnq/PZAEeZtBGiZdYxnIQhIOxV04nFmtOjSyyMt5NXMHFLHRnewSI6tojcFm2
        lm6WMZQXFqGyvkxjmPQXFFBCupAAPIkm0lIIvZqtvLOOqmU7et4Oy5pihyaHjg67
        cdmlv/p990e7YYfzdQSKezWrdwXW4EAGBTyuAZ67a5XCOUi17JZU
        -----END CERTIFICATE-----
        """
    }
}
